/*----------------------------------------------------------------------
  File    : mpr.c
  Contents: multivariate polynomial regression
  Author  : Christian Borgelt
  History : 2001.10.13 file created
            2001.10.15 first version completed
            2001.10.18 some messages added, option -v added
            2001.10.20 computation of sum of squared errors added
            2001.10.26 adapted to modified module regress
            2003.08.16 slight changes in error message output
            2003.10.25 logistic regression added (logit transformation)
            2007.02.13 adapted to modified module tabread
            2007.05.18 execution time measurements added
            2007.10.10 evaluation of attribute directions added
            2011.07.27 adapted to modified attset and utility modules
            2013.08.12 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.13 adapted to definition of type DIMID in matrix.h
            2013.08.29 adapted to new function as_target()
            2014.10.24 changed from LGPL license to MIT license
            2015.07.28 range for logit transform, two reg. parameters
            2015.07.31 forward selection of attributes based on AUC
            2015.08.03 adapted to changed regression object interface
            2015.08.05 balancing with integer factors made possible
            2015.09.11 error message "target attribute is not binary"
            2015.09.18 date and time output for attribute selection
            2015.11.30 nominal values mapped with 1-in-(n-1) encoding
            2018.09.13 extended functions (attset, attmap) removed
            2018.10.12 gradient methods and N-R for error function added
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>
#include "arrays.h"
#include "params.h"
#ifndef MAT_RDWR
#define MAT_RDWR
#endif
#include "matrix.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#include "attset.h"
#include "attmap.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_MATRIX
#define TAB_MATRIX
#endif
#include "table.h"
#include "regress.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "mpr"
#define DESCRIPTION "multivariate polynomial regression"
#define VERSION     "version 5.0 (2018.10.15)         " \
                    "(c) 2001-2018   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_TPLCNT    (-10)       /* data file is empty */
#define E_ATTCNT    (-11)       /* no usable attributes found */
#define E_INPCNT    (-12)       /* no usable inputs found */
#define E_UNKTRG    (-13)       /* unknown target attribute */
#define E_MULTRG    (-14)       /* multiple target attributes */
#define E_TRGBIN    (-15)       /* target attribute is not binary */
#define E_BALANCE   (-16)       /* unknown balancing mode */
#define E_DEGREE    (-17)       /* invalid degree of polynomial */
#define E_OPTMODE   (-18)       /* invalid optimization mode */
#define E_LOGIT     (-19)       /* logit transformation failed */
#define E_SOLVE     (-20)       /* failed to solve equation system */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*--------------------------------------------------------------------*/
#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

#if DIMID!=ATTID
#error "DIMID and ATTID must be the same data type"
#endif

#if DIMID!=TPLID
#error "DIMID and TPLID must be the same data type"
#endif

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- prediction for AUC comp. --- */
  double out;                   /* computed prediction */
  double trg;                   /* actual target value */
} PRED;                         /* (prediction pair) */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- mode information --- */
  int  code;                    /* code        of mode */
  char *name;                   /* name        of mode */
  char *desc;                   /* description of mode */
} MODEINFO;                     /* (mode information) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "data file is empty",
  /* E_ATTCNT  -11 */  "no (usable) attributes (need at least 1)",
  /* E_INPCNT  -12 */  "no (usable) inputs (need at least 1)",
  /* E_UNKTRG  -13 */  "unknown target attribute '%s'",
  /* E_MULTRG  -14 */  "multiple target attributes found",
  /* E_TRGBIN  -15 */  "target attribute '%s' is not binary",
  /* E_BALANCE -16 */  "unknown balancing mode %c",
  /* E_DEGREE  -17 */  "invalid degree of polynomial %d",
  /* E_OPTMODE -18 */  "invalid optimization mode %s",
  /* E_LOGIT   -19 */  "logit transformation failed for row %"DIMID_FMT,
  /* E_SOLVE   -20 */  "failed to solve equation system",
  /*           -21 */  "unknown error"
};

static const MODEINFO opttab[] = { /* table of optimization modes */
  { REG_NONE,      "none",
                   "no subsequent optimization"                       },
  { REG_ERR_GD,    "errgd",
                   "gradient descent on error function"               },
  { REG_ERR_GD,    "gd",                                         NULL },
  { REG_ERR_GD,    "d",                                          NULL },
  { REG_ERR_NR,    "errnr",
                   "Newton-Raphson   on error function"               },
  { REG_ERR_NR,    "enr",                                        NULL },
  { REG_ERR_NR,    "nre",                                        NULL },
  { REG_ERR_NR,    "e",                                          NULL },
  { REG_LLH_GA,    "llhga",
                   "gradient ascent  on likelihood function"          },
  { REG_LLH_GA,    "ga",                                         NULL },
  { REG_LLH_GA,    "a",                                          NULL },
  { REG_LLH_NR,    "llhnr",
                   "Newton-Raphson   on likelihood function"          },
  { REG_LLH_NR,    "lnr",                                        NULL },
  { REG_LLH_NR,    "nrl",                                        NULL },
  { REG_LLH_NR,    "l",                                          NULL },
  { -1,            NULL,        /* sentinel */                   NULL },
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan   = NULL;  /* scanner (for domain file) */
static TABREAD *tread  = NULL;  /* table reader */
static ATTSET  *attset = NULL;  /* attribute set */
static ATTMAP  *attmap = NULL;  /* attribute map */
static TABLE   *table  = NULL;  /* (training) data table */
static TABLE   *valid  = NULL;  /* validation data table */
static MATRIX  *data   = NULL;  /* vectors/data points as a matrix */
static double  *vec    = NULL;  /* vector to read table rows */
static REGRESS *reg    = NULL;  /* regression object */
static FILE    *out    = NULL;  /* output file */

/*----------------------------------------------------------------------
  Forward Selection of Attributes
----------------------------------------------------------------------*/

static int predcmp (const void *a, const void *b, void *data)
{                               /* --- compare two predictions */
  if (((PRED*)a)->out > ((PRED*)b)->out) return +1;
  if (((PRED*)a)->out < ((PRED*)b)->out) return -1;
  return 0;                     /* return sign of difference */
}  /* predcmp() */              /* of prediction values */

/*--------------------------------------------------------------------*/

static void printtime (void)
{                               /* --- report current date and time */
  time_t    t;                  /* current time */
  struct tm *tm;                /* current time as a time struct */
  char      buf[64];            /* for formating the current time */

  t  = time(NULL);              /* get the current time and */
  tm = localtime(&t);           /* turn it into a time struct */
  if (!tm) return;              /* check for correct conversion */
  strftime(buf, sizeof(buf)-1, "%F %T", tm);
  fprintf(stderr, "%s ", buf);  /* format and print current time */
}  /* printtime() */

/*--------------------------------------------------------------------*/

static int selerr (int errcode, REGRESS *reg, MATRIX *data,
                   ATTMAP *map, PRED *preds)
{                               /* --- clean up after sel. error */
  if (reg)   reg_delete(reg);   /* delete regression object, */
  if (data)  mat_delete(data);  /* created data matrix, */
  if (map)   am_delete(map, 0); /* attribute map and */
  if (preds) free(preds);       /* regression predictions */
  return     errcode;           /* return the error code */
}  /* selerr() */

/*--------------------------------------------------------------------*/

static DIMID fwdsel (TABLE *table, TABLE *valid, ATTID trgid, int deg,
                     int optmode, double logit[], double trpar[],
                     double lrate, double limit, int maxiter,
                     ATTID maxsel, int verbose)
{                               /* --- forward selection of atts. */
  ATTSET  *attset;              /* underlying attribute set */
  int     tt;                   /* target type */
  DIMID   j, k, m;              /* loop variables for attributes */
  DIMID   i, n;                 /* loop variables for tuples */
  DIMID   r;                    /* buffer for function results */
  ATT     *att;                 /* to traverse the attributes */
  TUPLE   *tpl;                 /* to traverse the tuples */
  INST    *col;                 /* to access the target column */
  TPLID   neg, pos, sum;        /* negative and positive counters */
  double  p;                    /* current prediction value */
  double  amin, amax, aavg;     /* area under the ROC curve (AUC) */
  double  bmin, bmax, bavg;     /* best minimum/maximum/average AUC */
  ATTID   bid;                  /* index of best attribute */
  ATTMAP  *attmap = NULL;       /* attribute map for conversion */
  MATRIX  *data   = NULL;       /* data matrix created from table */
  REGRESS *reg    = NULL;       /* induced regression object */
  PRED    *preds  = NULL;       /* regression predictions */
  double  *vec;                 /* vactor for validation mapping */
  clock_t t;                    /* for time measurements */

  assert(table                  /* check the function arguments */
  &&     valid && (trgid >= 0) && (deg >= 0));
  if (deg <= 0) return 0;       /* check for a positive degree */
  attset = tab_attset(table);   /* get the underlying attribute set */
  attmap = am_create(attset, 0, 1.0);
  if (!attmap) return E_NOMEM;  /* create an attribute map */
  am_target(attmap, trgid);     /* and set the target attribute */
  k = am_inpcnt(attmap);        /* get maximal number of inputs */
  am_delete(attmap, 0);         /* delete the attribute map */
  attmap = NULL;                /* and clear the variable */
  m = as_attcnt(attset);        /* get the number of attributes */
  as_setmark(attset, -1);       /* unmark all attributes */
  att = as_att(attset, trgid);  /* get and specially mark */
  att_setmark(att, 0);          /* the target attribute */
  tt = att_type(att);           /* get the target type */
  n  = tab_tplcnt(valid);       /* get number of validation tuples */
  preds = (PRED*)malloc((size_t)n *sizeof(PRED)
                       +(size_t)k *sizeof(double));
  if (!preds) return E_NOMEM;   /* create a prediction array and */
  vec = (double*)(preds +n);    /* a vector for validation mapping */
  t = clock();                  /* start timer, print log message */
  fprintf(stderr, "selecting attributes ...\n");
  bmin = bmax = bavg = -1.0;    /* clear the best attribute value */
  for (k = 0; k < m-1; k++) {   /* forward selection of attributes */
    bid = -1;                   /* clear the best attribute */
    for (j = 0; j < m; j++) {   /* traverse the attributes */
      att = as_att(attset, j);  /* skip target and selected attribs. */
      if (att_getmark(att) >= 0) continue;
      att_setmark(att, 1);      /* mark current attribute as selected */

      /* --- create attribute map and data matrix --- */
      attmap = am_create(attset, AM_MARKED, 1.0);
      if (!attmap) return selerr(E_NOMEM, reg, data, attmap, preds);
      am_target(attmap, trgid); /* create an attribute map and */
      r = am_inpcnt(attmap);    /* get the number of inputs */
      data = tab_matrix(table, attmap, AM_BOTH, 0);
      if (!data)   return selerr(E_NOMEM, reg, data, attmap, preds);

      /* --- build regression model --- */
      reg = reg_create(r+1,deg);/* create a regression object */
      if (!reg)    return selerr(E_NOMEM, reg, data, attmap, preds);
      reg_minmax(reg, logit[0], logit[1]);  /* set range and */
      reg_reglim(reg, trpar[0], trpar[1]);  /* regularization */
      reg_init(reg);            /* initialize the regression object */
      if (optmode != REG_NONE){ /* if optimized logistic regression */
        r = reg_logopt(reg, data, optmode, lrate, limit, maxiter);
        if (r > 0) return selerr(r, reg, data, attmap, preds); }
      else {                    /* if to do standard regression */
        r = reg_aggmat(reg, data);
        if (r > 0) return selerr(r, reg, data, attmap, preds);
        r = reg_solve(reg);     /* aggregate the data matrix */
      }                         /* solve the equation system */
      if (r < 0) {              /* check for an error */
        selerr(0, reg, data, attmap, NULL);
        att_setmark(att, -1); continue;
      }                         /* mark and skip attribute */

      /* --- evaluate model on validation data --- */
      for (i = 0; i < n; i++) { /* traverse the validation tuples */
        tpl = tab_tpl(valid, i);/* (assess the model performance) */
        col = tpl_colval(tpl, trgid);
        tpl_toas(tpl);          /* map tuple to numeric vector */
        am_exec(attmap, AM_INPUTS, vec);
        preds[i].trg = (tt == AT_FLT) ? (double)col->f
                     : (tt == AT_INT) ? (double)col->i : (double)col->n;
        preds[i].out = reg_exec(reg, vec);
      }                         /* collect target and prediction */
      obj_qsort(preds, (size_t)n, sizeof(PRED), +1, predcmp, NULL);
      neg  = pos  = sum = 0;    /* initialize the case counters */
      amin = amax = 0;          /* and the area under the ROC curve */
      p = preds[0].out;         /* get the first prediction */
      for (i = n; --i >= 0; ) { /* traverse the cases (in reverse) */
        if (preds[i].out != p){ /* if the prediction changes */
          p     = preds[i].out; /* note the new prediction value */
          amin += (double)neg *(double)sum;
          sum  += pos;          /* add current rectangle to AUC */
          amax += (double)neg *(double)sum;
          neg   = pos = 0;      /* reinitialize the counters */
        }                       /* (for next prediction value) */
        if (preds[i].trg > 0) pos += 1;
        else                  neg += 1;
      }                         /* update the case counters */
      amin += (double)neg *(double)sum;
      sum  += pos;              /* add last rectangle to AUC */
      amax += (double)neg *(double)sum;
      amin /= p = (double)sum *(double)(n -sum);
      amax /= p;                /* normalize by rectangle area */
      amin = 2.0 *amin -1.0;    /* compute area above diagonal */
      amax = 2.0 *amax -1.0;    /* (minimum/maximum/average) */
      aavg = 0.5 *amin +0.5 *amax;
      if (verbose) {            /* if verbose reporting requested */
        printtime();            /* print the current date and time */
        fprintf(stderr, "% .6f [% .6f,% .6f] %s\n",
                aavg, amin, amax, att_name(att));
      }                         /* print minimum and maximum AUC */
      if ((aavg > bavg) || (maxsel < -2)) {  /* update best att. */
        bavg = aavg; bmin = amin; bmax = amax; bid = j; }

      /* --- clean up variables --- */
      selerr(0, reg, data, attmap, NULL);
      att_setmark(att, -1);     /* unmark the current attribute */
      if (maxsel < -1) break;   /* abort the loop if to select */
    }                           /* the attributes in a fixed order */
    if (bid < 0) break;         /* if no attribute selected, abort */
    printtime();                /* print the current date and time */
    att = as_att(attset, bid);  /* get and report the best attribute */
    fprintf(stderr, "%3"ATTID_FMT": % .6f [% .6f,% .6f] %s\n",
            k+1, bavg, bmin, bmax, att_name(att));
    att_setwgt (att, (WEIGHT)bavg);   /* note AUC value in weight and */
    att_setmark(att, k+1);      /* mark best attribute with sel. step */
    if ((maxsel >= 0) && (k >= maxsel-1)) { k++; break; }
  }                             /* check for maximum number of atts. */
  if ((maxsel > -3)             /* if to do an actual selection */
  &&  (bavg   <= 0)) {          /* and result no better than random */
    as_setmark(attset, -1);     /* remove all selected attributes */
    att_setmark(as_att(attset, trgid), 0);
  }
  free(preds);                  /* delete the prediction array */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)] ", k);
  fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));
  return 0;                     /* return 'ok' */
}  /* fwdsel() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (reg)    reg_delete(reg);      \
  if (vec)    free(vec);            \
  if (data)   mat_delete(data);     \
  if (valid)  tab_delete(valid, 0); \
  if (table)  tab_delete(table, 0); \
  if (attmap) am_delete(attmap, 0); \
  if (attset) as_delete(attset);    \
  if (tread)  trd_delete(tread, 1); \
  if (scan)   scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static int read_table (ATTSET *attset, TABREAD *tread,
                       CCHAR *fn_tab, CCHAR *fn_hdr, int mode,
                       TABLE **tab)
{                               /* --- read a table file */
  int     k;                    /* return value of read function */
  ATTID   m;                    /* number of attributes */
  TPLID   n;                    /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* for time measurements */

  /* --- read table header --- */
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0) return E_FOPEN;
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) return -k;       /* read table header */
    trd_close(tread);           /* and close the file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0) return E_FOPEN;
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  *tab = tab_create("table", attset, tpl_delete);
  if (!*tab) return E_NOMEM;    /* create a data table */
  k = tab_read(*tab, tread, mode);
  if (k < 0) return -k;         /* read the table file */
  trd_close(tread);             /* and close the file */
  m = tab_attcnt(*tab);         /* get the number of attributes */
  n = tab_tplcnt(*tab);         /* and the number of data tuples */
  w = tab_tplwgt(*tab);         /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  return 0;                   /* return 'ok' */
}  /* read_table() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counter */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domain file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_val  = NULL;      /* name of validation data file */
  CCHAR   *fn_reg  = NULL;      /* name of regression function file */
  CCHAR   *blanks  = NULL;      /* blanks */
  CCHAR   *fldseps = NULL;      /* field  separators */
  CCHAR   *recseps = NULL;      /* record separators */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *trgname = NULL;      /* target attribute name */
  CCHAR   *optname = "none";    /* name of optimization method */
  int     matinp   = 0;         /* flag for numerical matrix input */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL; /* table read mode */
  int     onfly    = 0;         /* flag for "on the fly" processing */
  int     balance  = 0;         /* flag for balancing class freqs. */
  int     deg      = 1;         /* degree of polynomial */
  int     optmode  = REG_NONE;  /* logistic reg. optimization mode */
  double  lrate    = 0.05;      /* learning rate for gradient methods */
  int     maxiter  = 100;       /* maximum number of iterations */
  double  limit    = 1e-6;      /* limit for termination */
  double  logit[]  = {0,0};     /* scale for logit transformation */
  double  trpar[]  = {0,0};     /* Tikhonov regularization params. */
  ATTID   maxsel   = -1;        /* maximum number of atts. to select */
  int     verbose  = 0;         /* flag for verbose reporting */
  int     desc     = REG_TITLE; /* description mode */
  ATTID   trgid;                /* id of the target column */
  ATT     *target;              /* target attribute */
  ATTID   *perm;                /* permutation of attributes */
  ATTID   a;                    /* number of attributes */
  DIMID   m;                    /* number of variables/dimensions */
  DIMID   n;                    /* number of data points/vectors */
  double  w, u;                 /* weight of data points/vectors */
  DIMID   r, p;                 /* loop variable, buffer */
  clock_t t;                    /* for time measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] [-M|domfile [-d|-h hdrfile]] "
                  "tabfile regfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-M       input is a numerical matrix            "
                    "(default: input is a table)\n");
    printf("-c#      output/target attribute name           "
                    "(default: last attribute)\n");
    printf("-q#      balance class frequencies (weight tuples)\n");
    printf("         (l: lower, b: boost, s: shift tuple weights;\n");
    printf("          uppercase letters: use integer factors)\n");
    printf("         (only for nominal targets with two values)\n");
    printf("-x#      degree of regression polynomial        "
                    "(default: %d, i.e., linear)\n", deg);
    printf("-l#[:#]  range for logit transformation         "
                    "(default: normal regression)\n");
    printf("         (lower bound is 0 if only one value is given)\n");
    printf("-R#[:#]  Tikhonov regularization parameter      "
                    "(default: %g:%g)\n", trpar[0], trpar[1]);
    printf("         (negative first value: use only on failure;\n");
    printf("          second value limits regularization loop)\n");
    printf("-a#      logistic regression optimization mode  "
                    "(default: none)\n");
    for (i = 0; opttab[i].name; i++)
      if (opttab[i].desc)
        printf("         %-6s  %s\n", opttab[i].name, opttab[i].desc);
    printf("-t#      learning rate for optimization         "
                    "(default: %g)\n", lrate);
    printf("-T#      error change limit for termination     "
                    "(default: %g)\n", limit);
    printf("-i#      maximum number of iterations           "
                    "(default: %d)\n", maxiter);
    printf("         (-o/t/T/i only for logistic regression)\n");
    printf("-F       work \"on the fly\" if possible          "
                    "(default: load data)\n");
    printf("-v#      file name of validation data set       "
                    "(default: none)\n");
    printf("         (for a forward selection of attributes,\n");
    printf("         based on AUC, for two-class problems)\n");
    printf("-m#      maximum number of attributes to select "
                    "(default: no limit)\n");
    printf("-V       print verbose progress messages\n");
    printf("-e       print exponents of regression variables\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n"
           "         (only applicable with proper data table)\n");
    printf("domfile  file containing domain descriptions\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header  "
                    "(field names) from hdrfile\n");
    printf("hdrfile  file containing table header "
                    "(field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("regfile  file to write the regression model to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse the arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'M': matinp  = 1;                     break;
          case 'c': optarg  = &trgname;              break;
          case 'q': balance = (*s) ? *s++ : 0;       break;
          case 'x': deg     = (int)strtol(s, &s, 0); break;
          case 'l': getdblvec(s, &s, 2, logit);      break;
          case 'a': optarg  = &optname;              break;
          case 't': lrate   =      strtod(s, &s);    break;
          case 'R': getdblvec(s, &s, 2, trpar);      break;
          case 'T': limit   =      strtod(s, &s);    break;
          case 'i': maxiter = (int)strtol(s, &s, 0); break;
          case 'F': onfly   = 1;                     break;
          case 'v': optarg  = &fn_val;               break;
          case 'm': maxsel  = (int)strtol(s, &s, 0); break;
          case 'V': verbose = 1;                     break;
          case 'e': desc   |= REG_EXPOS;             break;
          case 'r': optarg  = &recseps;              break;
          case 'f': optarg  = &fldseps;              break;
          case 'b': optarg  = &blanks;               break;
          case 'C': optarg  = &comment;              break;
          case 'n': mode   |= AS_WEIGHT;             break;
          case 'd': mode   |= AS_DFLT;               break;
          case 'h': optarg  = &fn_hdr;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_reg = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note file names */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (matinp) {                 /* if input is a numerical matrix */
    if (k != 2) error(E_ARGCNT);/* check number of arguments */
    fn_reg = fn_tab;            /* shift the file names */
    fn_tab = fn_dom; fn_dom = NULL; }
  else {                        /* if input is a data table */
    if (k != 3) error(E_ARGCNT);/* check number of arguments */
    if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
      fn_hdr = "";              /* convert "-" to "" for consistency */
    i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
    if  (!fn_dom || !*fn_dom) i++;
    if  (!fn_tab || !*fn_tab) i++;
    if (i > 1) error(E_STDIN);  /* stdin must not be used twice */
  }                             /* poly. degree must be positive */
  if ((        balance  !=  0)  && (tolower(balance) != 'l')
  &&  (tolower(balance) != 'b') && (tolower(balance) != 's'))
    error(E_BALANCE, balance);  /* check the balancing mode */
  if (deg < 0) error(E_DEGREE, deg);
  if (logit[0] > logit[1]) {    /* ensure a proper range */
    w = logit[0]; logit[0] = logit[1]; logit[1] = w; }
  for (i = 0; opttab[i].name; i++)
    if (strcmp(opttab[i].name, optname) == 0) break;
  if (!opttab[i].name) error(E_OPTMODE, optname);
  optmode = opttab[i].code;     /* get optimization method code */
  if (optmode != REG_NONE)      /* ensure consistent parameters */
    onfly = 0;                  /* (optimization requires loading) */
  fputc('\n', stderr);          /* terminate the startup message */

  if (!matinp) {                /* if input is a data table */
    /* --- parse domain descriptions --- */
    attset = as_create("domains", att_delete);
    if (!attset) error(E_NOMEM);/* create an attribute set */
    scan = scn_create();        /* create a scanner */
    if (!scan)   error(E_NOMEM);/* for the domain file */
    t = clock();                /* start timer, open input file */
    if (scn_open(scan, NULL, fn_dom) != 0)
      error(E_FOPEN, scn_name(scan));
    fprintf(stderr, "reading %s ... ", scn_name(scan));
    if ((as_parse(attset, scan, AT_ALL, 1) != 0)
    ||  !scn_eof(scan, 1))      /* parse domain descriptions */
      error(E_PARSE, scn_name(scan));
    scn_delete(scan, 1);        /* delete the scanner and */
    scan = NULL;                /* clear the scanner variable */
    a = as_attcnt(attset);      /* get the number of attributes */
    if (a < 1) error(E_ATTCNT); /* check for at least one attribute */

    /* --- get and check target attribute --- */
    trgid = as_target(attset, trgname, 1);
    if (trgid < 0) error(trgname ? E_UNKTRG : E_MULTRG, trgname);
    target = as_att(attset, trgid); /* get and check target attribute */
    if ((att_type(target) == AT_NOM) && (att_valcnt(target) > 2))
      error(E_TRGBIN, att_name(as_att(attset, trgid)));
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a success message */

  if (matinp) {                 /* if input is a numerical matrix */
    /* --- read data points/vectors --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    m = -1;                     /* number of inputs is unknown */
    if (!onfly) {               /* if to load data into memory */
      k = mat_readx(&data, tread, -1, m);
      n = mat_rowcnt(data);     /* read the whole data matrix */
      m = mat_colcnt(data); }   /* and get its dimensions */
    else {                      /* if to work "on the fly" */
      k = vec_readx(&vec, &m, tread);
      if (k) error(k, TRD_INFO(tread));
    }                           /* read first vector/data point */
    if (m < 1) error(E_INPCNT); /* check input dimensions */

    /* --- create regression object --- */
    reg = reg_create(m, deg);   /* create a regression object and */
    if (!reg) error(E_NOMEM);   /* set range and regularization */
    reg_minmax(reg, logit[0], logit[1]);
    reg_reglim(reg, trpar[0], trpar[1]);
    reg_init(reg);              /* initialize the regression object */
    if (onfly) {                /* if to work "on the fly" */
      for (n = 0; k == 0; n++){ /* vector/data point read loop */
        if (reg_aggvec(reg, vec, 1.0) != 0)
          error(E_LOGIT, n+1);  /* aggregate current data point */
        k = vec_read(vec, m, tread);
      }                         /* read the next data point */
      if (k < 0) error(k, TRD_INFO(tread));
    }                           /* check for a read error */
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    fprintf(stderr, "[%"DIMID_FMT" variable(s),", m);
    fprintf(stderr, " %"DIMID_FMT" vector(s)]",   n);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    if (n < 1) error(E_TPLCNT); /* check for at least one data point */

    /* --- compute regression function --- */
    t = clock();                /* start timer, print log message */
    fprintf(stderr, "computing regression ... ");
    if (optmode != REG_NONE) {  /* if optimized logistic regression */
      n = reg_logopt(reg, data, optmode, lrate, limit, maxiter);
      if (n) error((n < 0) ? E_SOLVE : E_LOGIT, n); }
    else {                      /* if to do standard regression */
      if (!onfly) {             /* if data was loaded into memory */
        n = reg_aggmat(reg, data);
        if (n > 0) error(E_LOGIT, n+1);
      }                         /* aggregate the data matrix */
      if (reg_solve(reg) != 0)  /* compute regression coefficients */
        error(E_SOLVE);         /* by solving the equation system */
    }
    fprintf(stderr, "[sse: %g] ", reg_sse(reg));
    fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));
  }  /* if (matinp) .. */

  else if (onfly) {             /* if table and on the fly operation */
    /* --- create attribute map --- */
    attmap = am_create(attset, 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    am_target(attmap, trgid);   /* and set the target attribute */

    /* --- create regression object --- */
    m   = am_inpcnt(attmap)+1;  /* get the number of variables */
    reg = reg_create(m, deg);   /* create a regression object and */
    if (!reg) error(E_NOMEM);   /* set range and regularization */
    reg_minmax(reg, logit[0], logit[1]);
    reg_reglim(reg, trpar[0], trpar[1]);
    reg_init(reg);              /* initialize the regression object */

    /* --- read table header --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* set the separator characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    if (fn_hdr) {               /* if a header file is given */
      t = clock();              /* start timer, open input file */
      if (trd_open(tread, NULL, fn_hdr) != 0)
        error(E_FOPEN, trd_name(tread));
      fprintf(stderr, "reading %s ... ", trd_name(tread));
      k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
      if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
      trd_close(tread);         /* read table header, close file */
      fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", a);
      fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
      mode &= ~(AS_ATT|AS_DFLT);/* print a success message and */
    }                           /* remove the attribute flag */

    /* --- process data tuples --- */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_tab) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, mode); /* read/generate table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    vec = (double*)malloc((size_t)m *sizeof(double));
    if (!vec) error(E_NOMEM);   /* create a data point/vector buffer */
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    for (w = 0, n = 0; k == 0; n++) {
      w += u = as_getwgt(attset);   /* sum the tuple weights */
      am_exec(attmap, AM_ALL, vec); /* convert to vector and */
      k = reg_aggvec(reg, vec, u);  /* aggregate the current tuple */
      if (k) error(E_LOGIT, n+1);
      k = as_read(attset, tread, mode);
    }                           /* try to read the next tuple */
    free(vec); vec = NULL;      /* delete the vector buffer */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* close the input file and */
    tread = NULL;               /* delete the table reader */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", a);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    if (w <= 0) error(E_TPLCNT);/* print a success message */

    /* --- compute regression function --- */
    t = clock();                /* start the timer */
    fprintf(stderr, "computing regression ... ");
    if (reg_solve(reg) != 0)    /* compute regression coefficients */
      error(E_SOLVE);           /* by solving the equation system */
    fprintf(stderr, "[sse: %g] ", reg_sse(reg));
    fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));
  }                             /* write log message with sse */

  else {                        /* if to load the data into memory */
    /* --- read data table(s) --- */
    tread = trd_create();       /* create a table reader and */
    if (!tread) error(E_NOMEM); /* configure the characters */
    trd_allchs(tread, recseps, fldseps, blanks, "", comment);
    k = read_table(attset, tread, fn_tab, fn_hdr, mode, &table);
    if (k == E_NOMEM) error(E_NOMEM);
    if (k == E_FOPEN) error(E_FOPEN, trd_name(tread));
    if (k >  0)       error(k, as_errmsg(attset, NULL, 0));
    if (fn_val) {               /* if validation data is given */
      k = read_table(attset, tread, fn_val, fn_hdr, mode, &valid);
      if (k == E_NOMEM) error(E_NOMEM);
      if (k == E_FOPEN) error(E_FOPEN, trd_name(tread));
      if (k >  0)       error(k, as_errmsg(attset, NULL, 0));
    }                           /* read the valdation data */
    trd_delete(tread, 1);       /* (for forward attribute selection) */
    tread = NULL;               /* and delete the table reader */

    /* --- reduce and balance table --- */
    t = clock();                /* start timer, print log message */
    fprintf(stderr, "reducing%s table ... ",
                    (balance) ? " and balancing" : "");
    n = tab_reduce(table);      /* reduce table for speed up */
    if (balance                 /* if the balance flag is set */
    && (att_type(as_att(attset, trgid)) == AT_NOM)) {
      k = (tolower(balance) == 'l') ? -2
        : (tolower(balance) == 'b') ? -1 : 0;
      i = (tolower(balance) != balance);
      w = tab_balance(table, trgid, k, NULL, i);
      if (w < 0) error(E_NOMEM);/* balance the class frequencies */
    }                           /* and get the new weight sum */
    fprintf(stderr, "[%"TPLID_FMT, n);
    w = tab_getwgt(table, 0, n);/* get the total tuple weight */
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

    /* --- forward selection of attributes --- */
    if (valid) {                /* if to select attributes */
      n = fwdsel(table, valid, trgid, deg, optmode, logit, trpar,
                 lrate, limit, maxiter, maxsel, verbose);
      if (n > 0) error(E_LOGIT, n);
      if (n < 0) error((int)n); /* forward attribute selection */
      /* The result of the attribute selection is represented  */
      /* by attribute markers in the underlying attribute set. */
      tab_delete(valid, 0);     /* delete the validation data */
      valid = NULL;             /* and clear the variable */
      perm  = (ATTID*)malloc((size_t)a *sizeof(ATTID));
      if (!perm) error(E_NOMEM);/* create a permutation array */
      for (r = m = a; --r >= 0; ) { /* traverse the attributes */
        p = att_getmark(as_att(attset, r));
        if      (p > 0) perm[p-1] = r;
        else if (p < 0) perm[--m] = r;
      }                         /* collect destinations of columns */
      perm[--m] = trgid;        /* store target attribute and */
      trgid = m;                /* get the new target index */
      tab_colperm(table, perm); /* move selected attributes */
      free(perm);               /* to the front of the table */
    }                           /* and delete permutation array */

    /* --- compute (final) regression function --- */
    t = clock();                /* start timer, print log message */
    fprintf(stderr, "computing %s%sregression ... ",
                    (valid) ?               "final "    : "",
                    (logit[0] < logit[1]) ? "logistic " : "");
    attmap = am_create(attset, (valid) ? AM_MARKED : 0, 1.0);
    if (!attmap) error(E_NOMEM);/* create an attribute map */
    am_target(attmap, trgid);   /* and set the target attribute */
    data = tab_matrix(table, attmap, AM_BOTH, 1);
    if (!data) error(E_NOMEM);  /* convert the data table */
    table = NULL;               /* into a (data) matrix */
    m   = am_inpcnt(attmap)+1;  /* get the number of inputs, */
    reg = reg_create(m, deg);   /* create a regression object and */
    if (!reg) error(E_NOMEM);   /* set range and regularization */
    reg_minmax(reg, logit[0], logit[1]);
    reg_reglim(reg, trpar[0], trpar[1]);
    reg_init(reg);              /* initialize the regression object */
    if (optmode != REG_NONE) {  /* if optimized logistic regression */
      n = reg_logopt(reg, data, optmode, lrate, limit, maxiter);
      if (n) error((n < 0) ? E_SOLVE : E_LOGIT, n); }
    else {                      /* if to do standard regression */
      n = reg_aggmat(reg, data);/* aggregate the data matrix */
      if (n > 0) error(E_LOGIT, n);
      if (reg_solve(reg) != 0)  /* compute regression coefficients */
        error(E_SOLVE);         /* by solving the equation system */
    }
    fprintf(stderr, "[sse: %g] ", reg_sse(reg));
    fprintf(stderr, "done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- write regression function --- */
  t = clock();                  /* start timer, open output file */
  if (fn_reg && (strcmp(fn_reg, "-") == 0)) fn_reg = "";
  if (fn_reg && *fn_reg) { out = fopen(fn_reg, "w"); }
  else                   { out = stdout; fn_reg = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_reg);
  if (!out) error(E_FOPEN, fn_reg);
  if (!matinp) {                /* if input is a data table */
    k = AS_TITLE|AS_MARKED|AS_IVALS;
    if (fn_val) k |= AS_WEIGHT; /* form description mode */
    if (as_desc(attset, out, k, 0) != 0)
      error(E_FWRITE, fn_reg);  /* describe attribute domains */
    fprintf(out, "\n");         /* leave one line empty */
  }
  if (reg_desc(reg, out, desc, 0) != 0)
    error(E_FWRITE, fn_reg);    /* describe the regression polynomial */
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_reg);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[%"DIMID_FMT" parameter(s)]", reg_cnt(reg));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
